@@ -58,19 +58,20 @@ static char *name = "";
 static char *strippedname = "";
 static char description[255];
 static char cloneurl[1024];
-static int haslicense, hasreadme, hassubmodules, hastagsgph;
+static int hastagsgph;
+static char *submodules;
+static char *licensefiles[] = { "HEAD:LICENSE", "HEAD:LICENSE.md", "HEAD:COPYING" };
+static char *license;
+static char *readmefiles[] = { "HEAD:README", "HEAD:README.md" };
+static char *readme;
 static long long nlogcommits = -1; /* < 0 indicates not used */
 
 /* cache */
 static git_oid lastoid;
-static char lastoidstr[GIT_OID_HEXSZ + 2]; /* id + newline + nul byte */
+static char lastoidstr[GIT_OID_HEXSZ + 2]; /* id + newline + NUL byte */
 static FILE *rcachefp, *wcachefp;
 static const char *cachefile;
 
-#ifndef USE_PLEDGE
-#define pledge(p1,p2) 0
-#endif
-
 /* format `len' columns of characters. If string is shorter pad the rest
  * with characters `pad`. */
 int
@@ -144,6 +145,7 @@ commitinfo_getstats(struct commitinfo *ci)
 {
 	struct deltainfo *di;
 	git_diff_options opts;
+	git_diff_find_options fopts;
 	const git_diff_delta *delta;
 	const git_diff_hunk *hunk;
 	const git_diff_line *line;
@@ -161,10 +163,20 @@ commitinfo_getstats(struct commitinfo *ci)
 	}
 
 	git_diff_init_options(&opts, GIT_DIFF_OPTIONS_VERSION);
-	opts.flags |= GIT_DIFF_DISABLE_PATHSPEC_MATCH;
+	opts.flags |= GIT_DIFF_DISABLE_PATHSPEC_MATCH |
+	              GIT_DIFF_IGNORE_SUBMODULES |
+	              GIT_DIFF_INCLUDE_TYPECHANGE;
 	if (git_diff_tree_to_tree(&(ci->diff), repo, ci->parent_tree, ci->commit_tree, &opts))
 		goto err;
 
+	if (git_diff_find_init_options(&fopts, GIT_DIFF_FIND_OPTIONS_VERSION))
+		goto err;
+	/* find renames and copies, exact matches (no heuristic) for renames. */
+	fopts.flags |= GIT_DIFF_FIND_RENAMES | GIT_DIFF_FIND_COPIES |
+	               GIT_DIFF_FIND_EXACT_MATCH_ONLY;
+	if (git_diff_find_similar(ci->diff, &fopts))
+		goto err;
+
 	ndeltas = git_diff_num_deltas(ci->diff);
 	if (ndeltas && !(ci->deltas = calloc(ndeltas, sizeof(struct deltainfo *))))
 		err(1, "calloc");
@@ -172,6 +184,7 @@ commitinfo_getstats(struct commitinfo *ci)
 	for (i = 0; i < ndeltas; i++) {
 		if (git_patch_from_diff(&patch, ci->diff, i))
 			goto err;
+
 		if (!(di = calloc(1, sizeof(struct deltainfo))))
 			err(1, "calloc");
 		di->patch = patch;
@@ -461,12 +474,15 @@ writeheader(FILE *fp, const char *title)
 	fprintf(fp, "[1|Refs|%s/refs.gph|server|port]\n", relpath);
 	if (hastagsgph)
 		fprintf(fp, "[1|Tags|%s/tags.gph|server|port]\n", relpath);
-	if (hassubmodules)
-		fprintf(fp, "[1|Submodules|%s/file/.gitmodules.gph|server|port]\n", relpath);
-	if (hasreadme)
-		fprintf(fp, "[1|README|%s/file/README.gph|server|port]\n", relpath);
-	if (haslicense)
-		fprintf(fp, "[1|LICENSE|%s/file/LICENSE.gph|server|port]\n", relpath);
+	if (submodules)
+		fprintf(fp, "[1|Submodules|%s/file/%s.gph|server|port]\n",
+		        relpath, submodules);
+	if (readme)
+		fprintf(fp, "[1|README|%s/file/%s.gph|server|port]\n",
+		        relpath, readme);
+	if (license)
+		fprintf(fp, "[1|LICENSE|%s/file/%s.gph|server|port]\n",
+		        relpath, license);
 	fputs("---\n", fp);
 }
 
@@ -553,6 +569,7 @@ printshowfile(FILE *fp, struct commitinfo *ci)
 	git_patch *patch;
 	size_t nhunks, nhunklines, changed, add, del, total, i, j, k;
 	char buf[256], filename[256], linestr[32];
+	int c;
 
 	printcommit(fp, ci);
 
@@ -571,6 +588,17 @@ printshowfile(FILE *fp, struct commitinfo *ci)
 	fputs("Diffstat:\n", fp);
 	for (i = 0; i < ci->ndeltas; i++) {
 		delta = git_patch_get_delta(ci->deltas[i]->patch);
+
+		switch (delta->status) {
+		case GIT_DELTA_ADDED:      c = 'A'; break;
+		case GIT_DELTA_COPIED:     c = 'C'; break;
+		case GIT_DELTA_DELETED:    c = 'D'; break;
+		case GIT_DELTA_MODIFIED:   c = 'M'; break;
+		case GIT_DELTA_RENAMED:    c = 'R'; break;
+		case GIT_DELTA_TYPECHANGE: c = 'T'; break;
+		default:                   c = ' '; break;
+		}
+
 		if (strcmp(delta->old_file.path, delta->new_file.path)) {
 			snprintf(filename, sizeof(filename), "%s -> %s",
 				delta->old_file.path, delta->new_file.path);
@@ -578,7 +606,7 @@ printshowfile(FILE *fp, struct commitinfo *ci)
 		} else {
 			utf8pad(buf, sizeof(buf), delta->old_file.path, 35, ' ');
 		}
-		fputs("  ", fp);
+		fprintf(fp, "  %c ", c);
 		gphtext(fp, buf, strlen(buf));
 
 		add = ci->deltas[i]->addcount;
@@ -657,10 +685,10 @@ writelogline(FILE *fp, struct commitinfo *ci)
 	else
 		fputs("                ", fp);
 	fputs("  ", fp);
-	utf8pad(buf, sizeof(buf), ci->summary ? ci->summary : "", 50, ' ');
+	utf8pad(buf, sizeof(buf), ci->summary ? ci->summary : "", 40, ' ');
 	gphlink(fp, buf, strlen(buf));
 	fputs("  ", fp);
-	utf8pad(buf, sizeof(buf), ci->author ? ci->author->name : "", 25, ' ');
+	utf8pad(buf, sizeof(buf), ci->author ? ci->author->name : "", 19, '\0');
 	gphlink(fp, buf, strlen(buf));
 	fprintf(fp, "|%s/commit/%s.gph|server|port]\n", relpath, ci->oid);
 }
@@ -703,6 +731,8 @@ writelog(FILE *fp, const git_oid *oid)
 		} else if (nlogcommits > 0) {
 			writelogline(fp, ci);
 			nlogcommits--;
+			if (!nlogcommits && ci->parentoid[0])
+				fprintf(fp, "%16.16s  More commits remaining [...]\n", "");
 		}
 
 		if (cachefile)
@@ -1050,7 +1080,7 @@ writerefs(FILE *fp)
 			/* print header if it has an entry (first). */
 			if (++count == 1) {
 				fprintf(fp, "%s\n", titles[j]);
-				fprintf(fp, "  %-20.20s", "Name");
+				fprintf(fp, "  %-32.32s", "Name");
 				fprintf(fp, "  %-16.16s", "Last commit date");
 				fprintf(fp, "  %s\n", "Author");
 			}
@@ -1058,7 +1088,7 @@ writerefs(FILE *fp)
 			name = git_reference_shorthand(r);
 
 			fputs("  ", fp);
-			utf8pad(buf, sizeof(buf), name, 20, ' ');
+			utf8pad(buf, sizeof(buf), name, 32, ' ');
 			gphlink(fp, buf, strlen(buf));
 			fputs("  ", fp);
 			if (ci->author)
@@ -1067,7 +1097,7 @@ writerefs(FILE *fp)
 				fputs("                ", fp);
 			fputs("  ", fp);
 			if (ci->author) {
-				utf8pad(buf, sizeof(buf), ci->author->name, 25, ' ');
+				utf8pad(buf, sizeof(buf), ci->author->name, 25, '\0');
 				gphlink(fp, buf, strlen(buf));
 			}
 			fputs("\n", fp);
@@ -1097,7 +1127,7 @@ err:
 void
 usage(char *argv0)
 {
-	fprintf(stderr, "%s [-b baseprefix] [-c cachefile] [-l commits] repodir\n", argv0);
+	fprintf(stderr, "%s [-b baseprefix] [-c cachefile | -l commits] repodir\n", argv0);
 	exit(1);
 }
 
@@ -1135,10 +1165,7 @@ main(int argc, char *argv[])
 			errno = 0;
 			nlogcommits = strtoll(argv[++i], &p, 10);
 			if (argv[i][0] == '\0' || *p != '\0' ||
-			    nlogcommits <= 0)
-				usage(argv[0]);
-			if (errno == ERANGE && (nlogcommits == LLONG_MAX ||
-			    nlogcommits == LLONG_MIN))
+			    nlogcommits <= 0 || errno)
 				usage(argv[0]);
 		}
 	}
@@ -1150,6 +1177,7 @@ main(int argc, char *argv[])
 
 	git_libgit2_init();
 
+#ifdef __OpenBSD__
 	if (cachefile) {
 		if (pledge("stdio rpath wpath cpath fattr", NULL) == -1)
 			err(1, "pledge");
@@ -1157,6 +1185,7 @@ main(int argc, char *argv[])
 		if (pledge("stdio rpath wpath cpath", NULL) == -1)
 			err(1, "pledge");
 	}
+#endif
 
 	if (git_repository_open_ext(&repo, repodir,
 		GIT_REPOSITORY_OPEN_NO_SEARCH, NULL) < 0) {
@@ -1212,17 +1241,24 @@ main(int argc, char *argv[])
 	hastagsgph = (access("tags.gph", F_OK) == 0);
 
 	/* check LICENSE */
-	haslicense = (!git_revparse_single(&obj, repo, "HEAD:LICENSE") &&
-		git_object_type(obj) == GIT_OBJ_BLOB);
-	git_object_free(obj);
+	for (i = 0; i < sizeof(licensefiles) / sizeof(*licensefiles) && !license; i++) {
+		if (!git_revparse_single(&obj, repo, licensefiles[i]) &&
+		    git_object_type(obj) == GIT_OBJ_BLOB)
+			license = licensefiles[i] + strlen("HEAD:");
+		git_object_free(obj);
+	}
 
 	/* check README */
-	hasreadme = (!git_revparse_single(&obj, repo, "HEAD:README") &&
-		git_object_type(obj) == GIT_OBJ_BLOB);
-	git_object_free(obj);
+	for (i = 0; i < sizeof(readmefiles) / sizeof(*readmefiles) && !readme; i++) {
+		if (!git_revparse_single(&obj, repo, readmefiles[i]) &&
+		    git_object_type(obj) == GIT_OBJ_BLOB)
+			readme = readmefiles[i] + strlen("HEAD:");
+		git_object_free(obj);
+	}
 
-	hassubmodules = (!git_revparse_single(&obj, repo, "HEAD:.gitmodules") &&
-		git_object_type(obj) == GIT_OBJ_BLOB);
+	if (!git_revparse_single(&obj, repo, "HEAD:.gitmodules") &&
+	    git_object_type(obj) == GIT_OBJ_BLOB)
+		submodules = ".gitmodules";
 	git_object_free(obj);
 
 	/* log for HEAD */
@@ -1231,7 +1267,7 @@ main(int argc, char *argv[])
 	writeheader(fp, "Log");
 
 	fprintf(fp, "%-16.16s  ", "Date");
-	fprintf(fp, "%-50.50s  ", "Commit message");
+	fprintf(fp, "%-40.40s  ", "Commit message");
 	fprintf(fp, "%s\n", "Author");
 
 	if (cachefile && head) {
